#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <cmath>
#include <ctime>

struct WaveFmtSubChunk
{
   char         id [4];
   int          size          : 32;
   int          format        : 16;
   unsigned int channels      : 16;
   unsigned int sampleRate    : 32;
   unsigned int byteRate     : 32;
   unsigned int blockAlign    : 16;
   unsigned int bitsPerSample : 16;
};

struct WaveDataSubChunk
{
   char id [4];
   int  size : 32;
};

unsigned int fileSize = 0;
unsigned char* fileData = 0;
WaveDataSubChunk *dataChunk = 0;
WaveFmtSubChunk *fmtChunk = 0;
unsigned int rightSample = 0;
unsigned int leftFixed = 0;
unsigned int rightFixed = 0;
unsigned int prog = 0;
const unsigned int progressSamples = 65;
char progressBar[progressSamples];
unsigned int lastProgress = 0;
bool forceSongSpecials[4];

const unsigned int headerSize = 44;
std::ofstream logfile;

bool openWaveFile(const char* fileName)
{
   bool result = false;
   if (fileName)
   {
      std::ifstream waveFile;
      waveFile.open(fileName,std::ifstream::binary|std::ifstream::in);
      if (waveFile.is_open())
      {
         result = true;
         static_cast<void>(waveFile.seekg(0, std::ifstream::end));
         fileSize = static_cast<unsigned int>(waveFile.tellg()); // get file size
         static_cast<void>(waveFile.seekg(0,std::ifstream::beg));
         fileData = new unsigned char [fileSize];
         static_cast<void>(waveFile.read ((char *)fileData, static_cast<int>(fileSize)));

         // now check the data we got from the file
         if (  (strncmp (reinterpret_cast<char*>(fileData), "RIFF", 4))
             ||(strncmp (reinterpret_cast<char*>(fileData + 8), "WAVE", 4)))
         {
            std::cout << "Not a WAVE file!" << std::endl;
            result = false;
         }
         // now check if it is CD audio
         fmtChunk = (WaveFmtSubChunk*)(fileData + 12);
         if (result && (fmtChunk->format != 1))
         {
            std::cout << "Invalid audio format!" << std::endl;
            result = false;
         }
         if (result &&
            (  (fmtChunk->channels != 2) // stereo
             ||(fmtChunk->sampleRate != 44100) // 44 KHz
             ||(fmtChunk->bitsPerSample != 16) // 16 bit
             ||(fmtChunk->byteRate != 176400) // 176400 bits/second
             ||(fmtChunk->blockAlign != 4)   // 4 bytes per frame
            ))
         {
            // no valid CD audio data
            std::cout << "No valid CD audio data!" << std::endl;
            result = false;
         }
         // now get the actual data chunk
         dataChunk = (WaveDataSubChunk *)(fileData + fmtChunk->size + 20);

         if (result && (strncmp (dataChunk->id, "data", 4)))
         {
            // no wave data found
            std::cout << "No wave data found" << std::endl;
            result = false;
         }
         if (result)
         {
            // try detecting the actual song
            if (fileSize > 105000000 && fileSize < 110000000) forceSongSpecials[0] = true;
            else
            if (fileSize >  90000000 && fileSize <  95000000) forceSongSpecials[1] = true;
            else
            if (fileSize >  75000000 && fileSize <  85000000) forceSongSpecials[2] = true;
            else
            if (fileSize > 125000000 && fileSize < 135000000) forceSongSpecials[3] = true;
            else
            {
               std::cout << "File does not seem to be from CD Frumpy 2!" << std::endl;
               result = false;
            }
         }
         if (!result)
         {
            delete[] fileData;
         }
      }
      else std::cout << "Could not open file " << fileName << "!" << std::endl;
   }
   return result;
}

void calcTimeFromFrame(unsigned int frame, unsigned int& minutes, float& seconds)
{
   seconds = static_cast<float>(frame) / 44100;
   minutes = static_cast<unsigned int>(seconds)/60;
   seconds -= static_cast<float>(minutes)*60.f;
}

int getSample(unsigned int frame)
{
   unsigned int position = headerSize + frame*fmtChunk->blockAlign + rightSample*(fmtChunk->blockAlign/2);
   int sample = fileData[position] + (fileData[position+1] << 8);
   if (sample > 32767) sample -= 65536;
   return sample;
}

void fixSample(unsigned int frame)
{
   unsigned int position = headerSize + frame*fmtChunk->blockAlign + rightSample*(fmtChunk->blockAlign/2);
   int sample = getSample(frame);
   sample += 16384;
   if (sample < 0) sample += 65536;
   fileData[position+1] = static_cast<unsigned char>(static_cast<unsigned int>(sample) >> 8);
   fileData[position] = static_cast<unsigned char>(sample - (fileData[position+1] << 8));
   if (rightSample) rightFixed++;
   else leftFixed++;
}

void writeNewFile(char* fileName)
{
   if (fileName)
   {
      char* file = strtok(fileName,".");
      strcat(file,"_fixed.wav");
      std::ofstream waveFile;
      waveFile.open(file,std::ios::binary|std::ios::trunc);
      if (waveFile.is_open()) 
      {
         static_cast<void>(waveFile.write((char *)fileData,static_cast<int>(fileSize)));
         waveFile.close();
      }
      else std::cout << "ERROR writing new file!" << std::endl;
   }   
}

unsigned int calcProgress(unsigned int frame, unsigned int total)
{
   return frame*progressSamples / total;
}

void showKeepAlive()
{
   switch(prog)
   {
   case 0: std::cout << "\r/ ";break;
   case 1: std::cout << "\r- ";break;
   case 2: std::cout << "\r\\ ";break;
   case 3: std::cout << "\r| ";break;
   default: break;
   }
   prog = (prog + 1) % 4;
}

void showProgress(unsigned int progress)
{
   showKeepAlive();
   if (leftFixed + rightFixed > lastProgress)
   {
      lastProgress = leftFixed+rightFixed;
      progressBar[progress] = 88;
   }
   else progressBar[progress] = 111;
   std::cout << progressBar;
   std::cout << " " << leftFixed << "/" << rightFixed;
   std::cout << "\r";
}

void fixSampleAndLog(unsigned int frame,unsigned int count)
{
   unsigned int minutes;
   float seconds;
   calcTimeFromFrame(frame,minutes,seconds);
   logfile << count << " suspicious sample(s) in ";
   if (rightSample)
   {
      logfile << "right channel at ";
   }
   else
   {
      logfile << "left  channel at ";
   }
   logfile << minutes << ":" << std::fixed << seconds << ", frame " << frame << std::endl;
   for (unsigned int i = 0; i < count; i++)
   {
      fixSample(frame+i);
   }
}

unsigned int differenceDetection(unsigned int frame)
{
   int value[7];
   int diff[6];
   int diff2[5];
   for (unsigned int i = 0; i < 7; i++) value[i] = getSample(frame+i-3);
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 6; i++)   diff[i] = value[i+1]-value[i];
   for (unsigned int i = 0; i < 5; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }
   if (((value[3]+4000 < value[2])||(value[3]+3000 < value[2] && forceSongSpecials[2] && value[3] < -32000)) &&
       ((forceSongSpecials[2] && value[3]+5000 < value[4])||
        (!forceSongSpecials[2] && value[3]+6000 < value[4])
       )
       &&
       value[3] < 16384 &&
       diffSumBefore > 50000 &&
       (!forceSongSpecials[2] || diff[5] > -14450 || diff[5] < -14460 ||diff[2] < -25400 || diff[1] > 12600)
      )
   {      
      value[3] += 16384;
      for (unsigned int i = 2; i < 4; i++)   diff[i] = value[i+1]-value[i];
      for (unsigned int i = 1; i < 4; i++)   diff2[i] = diff[i+1]-diff[i];
      
      int diffSumAfter = 0;
      for (unsigned int i = 0; i < 5; i++)   diffSumAfter += abs(diff2[i]);
      if ((diffSumAfter*2 < diffSumBefore + 4000)||
          (diffSumBefore > 113000 && diffSumAfter < 70000)||
          (forceSongSpecials[2] && diffSumBefore > 68400 && diffSumAfter < 57300 && diffSumAfter > 57200)||
          (forceSongSpecials[2] && diffSumBefore > 73400 && diffSumAfter < 44700 && diffSumAfter > 44600)
         )
      {
         fixSampleAndLog(frame,1);
      }
   }
   return diffSumBefore;
}

unsigned int differenceDetectionDouble(unsigned int frame)
{
   int value[8];
   for (unsigned int i = 0; i < 8; i++)   value[i] = getSample(frame+i-3);
   int diff[7];
   int diff2[6];
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 7; i++)   diff[i] = value[i+1]-value[i];
   for (unsigned int i = 0; i < 6; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }   
   if (value[3]+10000 < value[2] &&
       value[4]+11000 < value[5] &&
       value[3] < 16384 && 
       value[4] < 16384 &&
       diffSumBefore > 55000 &&
       (!forceSongSpecials[2] || diff[4] > 23200 || diff[4] < 23150) &&
       (!forceSongSpecials[2] || diff[4] > 12250 || diff[4] < 12200) &&
       (!forceSongSpecials[2] || diff[4] > 11950 || diff[4] < 11900) &&
       (!forceSongSpecials[2] || diff[4] > 23400 || diff[4] < 23350)
      )
   {
      value[3] += 16384;
      value[4] += 16384;
      for (unsigned int i = 2; i < 5; i++)   diff[i] = value[i+1]-value[i];
      for (unsigned int i = 1; i < 5; i++)   diff2[i] = diff[i+1]-diff[i];
      if (forceSongSpecials[2] && ((diff2[5] < -15700 && diff2[5] > -15750)||(diff2[5] < -16400 && diff2[5] > -16450))) diff2[5] += 16384;
      
      int diffSumAfter = 0;
      for (unsigned int i = 0; i < 6; i++)   diffSumAfter += abs(diff2[i]);
      if ( diffSumAfter*6 < diffSumBefore ||
           diffSumAfter < 10000 ||
          (diffSumAfter*5 < diffSumBefore && abs(diff2[0]) < 1800 && abs(diff2[5]) < 1800) ||
          (diffSumAfter*5 < diffSumBefore && abs(diff[1])-abs(diff[2]) < 300 && abs(diff[3]) - abs(diff[4]) < 850) ||
          (diffSumAfter*5 < diffSumBefore && abs(diff2[1])-abs(diff2[2]) < 100 && abs(diff2[2])-abs(diff2[3]) < 1500) ||
          (diffSumAfter*23/5 < diffSumBefore && abs(diff2[0]) - abs(diff2[1]) < 350
                                             && abs(diff2[1])-abs(diff2[2]) < 170 
                                             && abs(diff2[3])-abs(diff2[4]) < 850) ||
          (diffSumAfter > 36800 && diffSumBefore > 99900 && diffSumAfter < diffSumAfter)
         )
      {
         fixSampleAndLog(frame,2);
      }
   }
   return diffSumBefore;
}

// fix last remaining drops across two samples - last pass
unsigned int differenceDetectionSpecialGypsy(unsigned int frame)
{
   int value[4];
   int diff[3];
   int diff2[2];
   for (unsigned int i = 0; i < 4; i++) value[i] = getSample(frame+i-1);
   for (unsigned int i = 0; i < 3; i++) diff[i] = value[i+1] - value[i];
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 2; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }   
   if (((diff[0] < -17500 && diff[2] > 10300)||
        (diff[0] < -15800 && diff[2] > 16200)||
        (diff[0] < -11500 && diff[2] > 21300)||
        (diff[0] < -12800 && diff[2] > 11500 && diff[1] > 500 && diff[1] < 550)||
        (diff[0] < -14400 && diff[2] > 15600 && diff[1] > -650 && diff[1] < -600)||
        (diff[0] < -12600 && diff[2] > 15800 && diff[1] > 1900 && diff[1] < 1950)||
        (diff[0] < -12300 && diff[2] > 12000 && diff[1] > 1250 && diff[1] < 1300 && value[1] > 0)||
        (diff[0] < -16300 && diff[2] > 14800 && diff[1] > -150 && diff[1] < -100)||
        (diff[0] < -15200 && diff[2] > 14800 && diff[1] > 1400 && diff[1] < 1450)
       )&&
       abs(value[1]-value[2]) < 8500 &&
       value[1] < 16384          &&
       value[2] < 16384)
   {
      // corrected sum would be better, now check if there's a steep drop to the left or right
      // if there is, this drop is part of a longer series and shall not be fixed
      int valueRight[5];
      int diffRight[4];
      for (unsigned int i = 0; i < 5; i++) valueRight[i] = getSample(frame+i+2);
      for (unsigned int i = 0; i < 4; i++) diffRight[i] = -(valueRight[i]-valueRight[i+1]);
      if (diffRight[0] > -8100 &&
          diffRight[1] > -8100 &&
          diffRight[2] > -8100 &&
          diffRight[3] > -8100 &&
          diffRight[3] < 14600
         )
      {
         fixSampleAndLog(frame,2);
      }
      else
      {
         unsigned int absDiffSumRight = 0;
         for (unsigned int i = 0; i < 4; i++) absDiffSumRight += abs(diffRight[i]);
         if (absDiffSumRight > 36500)
         {
            // check if there is another drop to the right
            // repeat check from above
            int value[7];
            for (unsigned int i = 0; i < 7; i++) value[i] = getSample(frame+i-1);
            if (((value[4]+17500 < value[3] && value[5]+10300 < value[6])||
                 (value[4]+15800 < value[3] && value[5]+16200 < value[6])||
                 (value[4]+11500 < value[3] && value[5]+21300 < value[6])
                )&&
                abs(value[4]-value[5]) < 8500 &&
                value[4] < 16384          &&
                value[5] < 16384)
            {
               int diff[6];
               unsigned int diffSum = 0;
               for (unsigned int i = 0; i < 6; i++) 
               {
                  diff[i] = value[i]-value[i+1];
                  diffSum += abs(diff[i]);
               }
               value[1] += 16384;
               value[2] += 16384;
               value[4] += 16384;
               value[5] += 16384;
               unsigned int diffSumAfter = 0;
               for (unsigned int i = 0; i < 6; i++) 
               {
                  diff[i] = value[i]-value[i+1];
                  diffSumAfter += abs(diff[i]);
               }
               if (diffSumAfter < diffSum)
               {
                  fixSampleAndLog(frame,2);
                  fixSampleAndLog(frame+3,2);
               }
            }
         }
      }
   }
   return diffSumBefore;
}

unsigned int differenceDetectionSpecialCare(unsigned int frame)
{
   int value[4];
   int diff[3];
   int diff2[2];
   for (unsigned int i = 0; i < 4; i++) value[i] = getSample(frame+i-1);
   for (unsigned int i = 0; i < 3; i++) diff[i] = value[i+1] - value[i];
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 2; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }   
   if (((diff[0] < -16400 && diff[2] > 17400 && diff[1] > 6300 && diff[1] < 6350)||
        (diff[0] < -16300 && diff[2] > -350 && diff[2] < -300 && diff[1] > -250 && diff[1] < -200)||
        (diff[0] < -16500 && diff[2] > 16000 && diff[1] > -150 && diff[1] < -100)||
        (diff[0] < -22800 && diff[2] > 21400)||
        (diff[0] < -16700 && diff[2] > 11500 && diff[1] > -2300 && diff[1] < -2100)||
        (diff[0] < -16300 && diff[2] > 14800 && diff[1] > -150 && diff[1] < -100)||
        (diff[0] < -19300 && diff[2] > 18200 && diff[1] > 2800 && diff[1] < 2850)
       )&&
       abs(value[1]-value[2]) < 6400 &&
       value[1] < 16384          &&
       value[2] < 16384)
   {
      // corrected sum would be better, now check if there's a steep drop to the left or right
      // if there is, this drop is part of a longer series and shall not be fixed
      int valueRight[5];
      int diffRight[4];
      for (unsigned int i = 0; i < 5; i++) valueRight[i] = getSample(frame+i+2);
      for (unsigned int i = 0; i < 4; i++) diffRight[i] = -(valueRight[i]-valueRight[i+1]);
      if (diffRight[0] > -8100 &&
          diffRight[1] > -8100 &&
          diffRight[2] > -8100 &&
          diffRight[3] > -8100 &&
          diffRight[2] < 14600
         )
      {
         fixSampleAndLog(frame,2);
      }
      else
      {
         // check for five dropped samples in a row or drop immediately behind
         unsigned int absDiffSumRight = 0;
         for (unsigned int i = 0; i < 4; i++) absDiffSumRight += abs(diffRight[i]);
         if (absDiffSumRight > 16300)
         {
            // check if there is drop of five consecutive samples
            // repeat check from above
            int value[7];
            for (unsigned int i = 0; i < 7; i++) value[i] = getSample(frame+i-1);
            if (value[1]+16300 < value[0] && 
                value[5]+16200 < value[6] &&                
                value[1] < -17700 &&
                value[2] < -17900 &&
                value[3] < -18200 &&
                value[4] < -18200 &&
                value[5] < -18200)
            {
               int diff[6];
               unsigned int diffSum = 0;
               for (unsigned int i = 0; i < 6; i++) 
               {
                  diff[i] = value[i]-value[i+1];
                  diffSum += abs(diff[i]);
               }
               value[1] += 16384;
               value[2] += 16384;
               value[3] += 16384;
               value[4] += 16384;
               value[5] += 16384;
               unsigned int diffSumAfter = 0;
               for (unsigned int i = 0; i < 6; i++) 
               {
                  diff[i] = value[i]-value[i+1];
                  diffSumAfter += abs(diff[i]);
               }
               if (diffSumAfter < diffSum)
               {
                  fixSampleAndLog(frame,5);
               }
            }
            else
            {
               if (value[1]+16500 < value[0] && 
                   value[5]+16100 < value[6] &&
                   value[2]+16000 < value[3] &&
                   value[4]+16400 < value[3] &&
                   value[1] < -18900 &&
                   value[2] < -19000 &&
                   value[4] < -19400 &&
                   value[5] < -19500)
               {
                  int diff[6];
                  unsigned int diffSum = 0;
                  for (unsigned int i = 0; i < 6; i++) 
                  {
                     diff[i] = value[i]-value[i+1];
                     diffSum += abs(diff[i]);
                  }
                  value[1] += 16384;
                  value[2] += 16384;
                  value[4] += 16384;
                  value[5] += 16384;
                  unsigned int diffSumAfter = 0;
                  for (unsigned int i = 0; i < 6; i++) 
                  {
                     diff[i] = value[i]-value[i+1];
                     diffSumAfter += abs(diff[i]);
                  }
                  if (diffSumAfter < diffSum)
                  {
                     fixSampleAndLog(frame,2);
                     fixSampleAndLog(frame+3,2);
                  }
               }
            }
         }
      }
   }
   return diffSumBefore;
}

unsigned int differenceDetectionSpecial(unsigned int frame)
{
   if (forceSongSpecials[1]) return differenceDetectionSpecialGypsy(frame);
   if (forceSongSpecials[2]) return differenceDetectionSpecialCare(frame);
   return 0;
}

unsigned int differenceDetectionTriple(unsigned int frame)
{
   int value[9];
   int diff[8];
   int diff2[7];
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 9; i++)   value[i] = getSample(frame+i-3);
   for (unsigned int i = 0; i < 8; i++)   diff[i] = value[i+1]-value[i];
   for (unsigned int i = 0; i < 7; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }
   if (diff[2] < -14000 &&
       diff[5] >  11000 &&
       value[3] < 16384 && 
       value[4] < 16384 && 
       value[5] < 16384)
   {
      value[3] += 16384;
      value[4] += 16384;
      value[5] += 16384;
      for (unsigned int i = 2; i < 6; i++)   diff[i] = value[i+1]-value[i];
      for (unsigned int i = 1; i < 6; i++)   diff2[i] = diff[i+1]-diff[i];
      
      int diffSumAfter = 0;
      for (unsigned int i = 0; i < 7; i++)   diffSumAfter += abs(diff2[i]);
      if ( diffSumAfter*4 < diffSumBefore ||
          (diffSumAfter*5/2 < diffSumBefore && abs(diff2[2]) < 1000
                                            && abs(diff2[3]) < 500
                                            && abs(diff2[4]) < 500)
         )
      {
         fixSampleAndLog(frame,3);
      }
   }
   return diffSumBefore;
}

unsigned int differenceDetectionQuad(unsigned int frame)
{
   int value[10];
   int diff[9];
   int diff2[8];
   int diffSumBefore = 0;
   for (unsigned int i = 0; i < 10; i++)   value[i] = getSample(frame+i-3);
   for (unsigned int i = 0; i < 9; i++)   diff[i] = value[i+1]-value[i];
   for (unsigned int i = 0; i < 8; i++)   
   {
      diff2[i] = diff[i+1]-diff[i];
      diffSumBefore += abs(diff2[i]);
   }
   if (diff[2]  < -15600 &&
       diff[6]  > 13900 &&
       value[3] < 16384 && 
       value[4] < 16384 && 
       value[5] < 16384 && 
       value[6] < 16384)
   {
      value[3] += 16384;
      value[4] += 16384;
      value[5] += 16384;
      value[6] += 16384;
      for (unsigned int i = 2; i < 7; i++)   diff[i] = value[i+1]-value[i];
      for (unsigned int i = 1; i < 7; i++)   diff2[i] = diff[i+1]-diff[i];
      
      int diffSumAfter = 0;
      for (unsigned int i = 0; i < 8; i++)   diffSumAfter += abs(diff2[i]);
      if (diffSumAfter*4 < diffSumBefore)
      {
         fixSampleAndLog(frame,4);
      }
   }
   return diffSumBefore;
}

int main(int argc, char *argv[])
{
   std::cout << "frumpair - wave repair for Frumpy 2, written by Uwe Domaratius" << std::endl;
   if (argc < 2)
   {
      std::cout << "Usage: frumpair filename.wav" << std::endl;
   }
   else
   if (openWaveFile(argv[1]))
   {
      time_t start,end;
      time(&start);
      for (unsigned int i = 0; i < progressSamples; i++) progressBar[i] = 46;
      // print length of file
      float seconds;
      unsigned int minutes;
      unsigned int numberOfFrames = static_cast<unsigned int>(dataChunk->size) / fmtChunk->blockAlign;
      calcTimeFromFrame(numberOfFrames, minutes, seconds);
      std::cout << "File size " << fileSize << " bytes, ";
      std::cout << "length " << minutes << ":" << std::fixed << seconds << " minutes" << std::endl;
      if (forceSongSpecials[1]) std::cout << "File treated as: How The Gypsy Was Born" << std::endl;
      if (forceSongSpecials[2]) std::cout << "File treated as: Take Care Of Illusion" << std::endl;
      if (forceSongSpecials[1] || forceSongSpecials[2])
      {
         // create logfile
         logfile.open ("log.txt");

         unsigned int progress = 0;
         unsigned int oldProgress = 1;
         unsigned int currentSumL = 0;
         unsigned int currentSumR = 0;
         unsigned int sum = 0;
         const unsigned int threshold[5] = { 12000, 16000, 33000, 30000, 12000};
         const unsigned int step[5] = { 4, 5, 5, 6, 3 };

         // now do five passes to correct any wrong data
         for (unsigned int pass = 0; pass < 5; pass++)
         {
            unsigned int frame = 3;
            for (unsigned int i = 0; i < progressSamples; i++) progressBar[i] = 46;
            while (frame < numberOfFrames)
            {
               // fix the sample in the left channel
               rightSample = 0;
               switch (pass)
               {
               case 0: currentSumL = differenceDetection(frame);break;
               case 1: currentSumL = differenceDetectionDouble(frame);break;
               case 2: currentSumL = differenceDetectionTriple(frame);break;
               case 3: currentSumL = differenceDetectionQuad(frame);break;
               case 4: currentSumL = differenceDetectionSpecial(frame);break;
               }
               
               rightSample = 1;
               switch (pass)
               {
               case 0: currentSumR = differenceDetection(frame);break;
               case 1: currentSumR = differenceDetectionDouble(frame);break;
               case 2: currentSumR = differenceDetectionTriple(frame);break;
               case 3: currentSumR = differenceDetectionQuad(frame);break;
               case 4: currentSumR = differenceDetectionSpecial(frame);break;
               }

               sum = currentSumL > currentSumR ? currentSumL : currentSumR;
               
               if (frame % 10000 == 0)
               {
                  showKeepAlive();
               }
               progress = calcProgress(frame,numberOfFrames);
               if (progress != oldProgress)
               {
                  showProgress(progress);
                  oldProgress = progress;
               }
               if (sum > threshold[pass]) frame++;
               else frame += step[pass];
            }
            std::cout << std::endl;
         }

         // close logfile and write corrected wave file
         std::cout << std::endl << leftFixed << " suspicious samples in left channel"<< std::endl;
         std::cout << rightFixed << " suspicious samples in right channel" << std::endl;
         std::cout << leftFixed+rightFixed << " samples fixed" << std::endl;
         logfile << std::endl << leftFixed << " suspicious samples in left channel"<< std::endl;
         logfile << rightFixed << " suspicious samples in right channel" << std::endl;
         logfile << leftFixed+rightFixed << " samples fixed" << std::endl;
         logfile.close();
         std::cout << "Writing fixed file..." << std::endl;
         writeNewFile(argv[1]);
         time(&end);
         std::cout << static_cast<int>(difftime(end,start)) << " seconds elapsed" << std::endl;
      }
      else
      {
         if (forceSongSpecials[0]) std::cout << "Good Winds does not have any incorrect samples!" << std::endl;
         else
            if (forceSongSpecials[3]) std::cout << "Duty does not have any incorrect samples!" << std::endl;
            else std::cout << "File does not seem to be from Frumpy 2!" << std::endl;
      }
      delete[] fileData;
   }
   return 0;   
}
